! Cross Gradient function
! Author:  Mohammad Maysami
!          SEP, Stanford University
!          Apr 09
program Xgrad_it

use sep 
use sep3d_struct_mod
use xgrad_mod
use dottest


implicit none
type(sep3d)       :: mod_s,xg_s,aux_s
real,allocatable  :: mod(:),xg(:)
real   , pointer  :: aux(:)
real   , dimension(:), pointer  :: vmt, udt, dt,mt
logical           :: adj,add,test
integer           :: ns(3),nt,stat,gflag
real              :: os(3),ds(3)
real,dimension(2) :: dot1, dot2
integer, external :: auxin,sreed,srite

call sep_init(SOURCE)
! call getch_add_string("head=/dev/null")
! call getch_add_string("noheader=y")
! call set_no_putch()
! call sep_begin_prog()


call from_param("test",test,.false.)

if (.not. test) then
   !   Normal XGRAD IT
   ! ----------------------
   call init_sep3d("aux",aux_s,"INPUT")
   call init_sep3d("in",mod_s,"INPUT")

   call from_param("adj",adj,.false.)
   call from_param("add",add,.false.)



   ! Check size of mod and aux input
   nt=product(mod_s%n(1:sep3d_ndims(mod_s)))
   if (nt/=product(aux_s%n(1:sep3d_ndims(aux_s)))) &
        call seperr("expecting mod and auxilary to be the same size")

  
   ! Allocate mod and xg
   allocate( aux(nt) ,mod(nt),xg(nt))
   mod=0
   xg=0
   
   ! Read in auxilary xg and initialize xg operator
   if(4*size(aux)/=sreed("aux",aux,size(aux)*4)) then
      write(0,*) "Trouble reading in aux !"
   end if
   
   call xgrad_init(aux_s,aux)
   
   ! Read in mode and write out the X-gradient
   if(4*size(mod)/=sreed("in",mod,size(mod)*4)) then
      write(0,*) "Trouble reading in mod !"
   end if

   stat=xgrad_lop(adj,add,mod,xg)
   if(4*size(xg)/=srite("out",xg,size(xg)*4)) then
      call seperr("trouble writing out xgrad model")
   end if
   call sep_close



else
   !   Run Dot Tests
   ! ===================
   call from_param("ns",ns,(/500,400,1/))
   call from_param("ds",ds,(/1.,1.,1./))
   call from_param("os",os,(/0.,0.,0./))
   write(0,*) "      ns: ",ns
   write(0,*) "      ds: ",ds
  

   call init_sep3d(xg_s,"OUTPUT","FLOAT","REGULAR",ns,os,ds)
   call sep3d_write_description("out",xg_s)

   nt=ns(1)*ns(2)*ns(3)
   if (ns(3)>1) then
      call to_history("n4",3)
      allocate(xg(nt*3))
   else
      allocate(xg(nt*1))
   end if
  allocate(mod(nt),aux(nt))
  call random_number(mod)
  call random_number(aux)
     
  call xgrad_init(xg_s,aux)

  ! ! ===== SEPLIB Dot Test =====
  write(0,*) " ==> SEPLIB DOT TEST on xgrad_lop ..."
  write(0,*) "      n3: ",ns(3)
  call dot_test( xgrad_lop, (nt), (nt*3), dot1, dot2)


  ! ===== Manual Dot Test 0 =====
  write(0,*) " ==> Manual DOT TEST 000 on xgrad ..."
  write(0,*) "      n3: ",ns(3)

  if (ns(3) > 1) then
     allocate(mt(nt*3))
     allocate(vmt(nt*3))
     
     allocate(dt(nt*3))
     allocate(udt(nt*3))
  else
     allocate(mt(nt*1))
     allocate(vmt(nt*1))
     
     allocate(dt(nt*3))
     allocate(udt(nt*3))
 end if
     
 call random_number(vmt)
 call random_number(udt)
 call xgrad_init(xg_s,aux)
 
 stat=xgrad_lop(.false.,.false.,vmt,dt)
 stat=xgrad_lop (.true.,.false.,mt,udt)  
 dot1(1) = dot_product(udt, dt)
 dot1(2) = dot_product(vmt, mt)
 
 stat=xgrad_lop(.false.,.true.,vmt,dt)
 stat=xgrad_lop (.true.,.true.,mt,udt)  
 dot2(1) = dot_product(udt, dt)
 dot2(2) = dot_product(vmt, mt)
     
 write(0,*) "dot1 (No add):",dot1 
 write(0,*) "dot2 (   add):",dot2

end if
   
end program 





!   ! ===== Manual Dot Test =====
!   write(0,*) " ==> Manual DOT TEST on grad/cross_lop ..."
!   write(0,*) "      n3: ",ns(3)

!   if (ns(3)>1) then
!      gflag=0
!      write(0,*) "      gflag:",gflag
!      if (gflag==1) then
!         allocate(mt(nt*1))
!         allocate(vmt(nt*1))
        
!         allocate(dt(nt*3))
!         allocate(udt(nt*3))
!      else
!         allocate(mt(nt*3))
!         allocate(vmt(nt*3))
        
!         allocate(dt(nt*3))
!         allocate(udt(nt*3))
!      end if
     
!      call random_number(vmt)
!      call random_number(udt)
!      call xgrad_init(xg_s,aux)

!      stat=cross_lop3(.false.,.false.,vmt,dt)
!      stat=cross_lop3(.true.,.false.,mt,udt)  
!      dot1(1) = dot_product(udt, dt)
!      dot1(2) = dot_product(vmt, mt)
     
!      stat=cross_lop3(.false.,.true.,vmt,dt)
!      stat=cross_lop3(.true.,.true.,mt,udt)  
!      dot2(1) = dot_product(udt, dt)
!      dot2(2) = dot_product(vmt, mt)

  
  
!   if (ns(3)>1) then
!      gflag=0
!      write(0,*) "      gflag:",gflag
!      if (gflag==1) then
!         allocate(mt(nt*1))
!         allocate(vmt(nt*1))
        
!         allocate(dt(nt*3))
!         allocate(udt(nt*3))
!      else
!         allocate(mt(nt*3))
!         allocate(vmt(nt*3))
        
!         allocate(dt(nt*3))
!         allocate(udt(nt*3))
!      end if
     
!      call random_number(vmt)
!      call random_number(udt)
!      call xgrad_init(xg_s,aux)
     
!      stat=cross_lop3(.false.,.false.,vmt,dt)
!      stat=cross_lop3(.true.,.false.,mt,udt)  
!      dot1(1) = dot_product(udt, dt)
!      dot1(2) = dot_product(vmt, mt)
     
!      stat=cross_lop3(.false.,.true.,vmt,dt)
!      stat=cross_lop3(.true.,.true.,mt,udt)  
!      dot2(1) = dot_product(udt, dt)
!      dot2(2) = dot_product(vmt, mt)
     
!   else
!      gflag=1
!      write(0,*) "      gflag:",gflag
!      if (gflag==1) then
!         allocate(mt(nt*1))
!         allocate(vmt(nt*1))
        
!         allocate(dt(nt*2))
!         allocate(udt(nt*2))
!      else
!         allocate(mt(nt*2))
!         allocate(vmt(nt*2))
        
!         allocate(dt(nt*1))
!         allocate(udt(nt*1))
!      end if
     
!      call random_number(vmt)
!      call random_number(udt)
!      call xgrad_init(xg_s,aux)
     
!      stat=grad_lop2(.false.,.false.,vmt,dt)
!      stat=grad_lop2(.true.,.false.,mt,udt)
     
!      dot1(1) = dot_product(udt, dt)
!      dot1(2) = dot_product(vmt, mt)
     
!      stat=grad_lop2(.false.,.true.,vmt,dt)
!      stat=grad_lop2(.true.,.true.,mt,udt)  
!      dot2(1) = dot_product(udt, dt)
!      dot2(2) = dot_product(vmt, mt)
!   end if
