! Cross Gradient function
! Author:  Mohammad Maysami
!          SEP, Stanford University
!          Apr 09
program test_xg

use sep 
use sep3d_struct_mod
use xgrad_mod
use igrad2
use helderiv

implicit none
type(sep3d)       :: mm_s,test_s,aux_s
real,allocatable  :: mm(:),test1(:),test2(:),aux(:)
integer           :: ns(3),nt,stat,typ
logical           :: adj,add
integer, external :: auxin,sreed,srite

call sep_init(SOURCE)


call from_param("type",typ,0)
call from_param("adj",adj,.false.)
call from_param("add",add,.true.)


!   Test Grad2d Op.
! ----------------------
call init_sep3d("in",mm_s,"INPUT")
nt=product(mm_s%n(1:sep3d_ndims(mm_s)))
ns=mm_s%n



! Allocate mm and test
allocate( mm(nt),test1(nt),test2(nt*2))


! Read in mode and write out the X-gradient
if(4*size(mm)/=sreed("in",mm,size(mm)*4)) then
   write(0,*) "Trouble reading in model !"
end if


if (typ==0) then 

   ! ! HELDERIV  Operator
   ! ! =====================
   write(0,*) " Entering Helix Derivative Mode"
   call helderiv_init(ns(1),32,0.0001 )
   stat=helderiv_lop( adj, add, mm, test1)
   
   if(4*size(test1)/=srite("out",test1,size(test1)*4)) then
      call seperr("trouble writing out test1")
   end if
   
else   
   if (typ==1) then
      ! IGRAD 2 Operator (similar 2 XGRAD)
      ! ==================================
      write(0,*) " Entering IGRAD2 Mode"
      call igrad2_init ( ns(1),ns(2) )
      stat=igrad2_lop ( adj, add, mm, test2)
   else      
      ! GRAD2 Operator 
      ! ==================================
      write(0,*) " Entering XGRAD Mode"
      call xgrad_init(mm_s,mm)
      nt=grad_lop2(adj,add,mm,test2)
   end if


   call to_history("n4",2)
   if(4*size(test2)/=srite("out",test2,size(test2)*4)) then
      call seperr("trouble writing out test2")
   end if

end if      
call sep_close

end program
